/*
 * GluonJ, An Aspect-Oriented Programming System for Java.
 * Copyright (C) 2006 Shigeru Chiba. All Rights Reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License.  Alternatively, the contents of this file may be used under
 * the terms of the GNU Lesser General Public License Version 2.1 or later.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 */
package javassist.gluonj.weave;

import java.util.HashMap;
import java.util.ArrayList;
import javassist.*;
import javassist.bytecode.AnnotationsAttribute;
import javassist.bytecode.ConstantAttribute;
import javassist.bytecode.FieldInfo;
import javassist.bytecode.AttributeInfo;
import javassist.bytecode.annotation.Annotation;
import javassist.gluonj.WeaveException;

/**
 * Refine class processor.
 */
public class RefineWeaver {
    /**
     * Returns the target class of the given refine class.
     */
    public CtClass getTarget(CtClass refiner) throws WeaveException {
        try {
            return getTarget0(refiner);
        }
        catch (NotFoundException nfe) {
            throw new WeaveException(nfe);
        }
    }

    private CtClass getTarget0(CtClass refiner) throws WeaveException, NotFoundException {
        boolean isIntf = refiner.isInterface();
        if (isIntf) {
            CtClass[] interfaces = refiner.getInterfaces();
            if (interfaces.length > 0)
                return interfaces[0];
            else
                throw new WeaveException(refiner.getName()
                            + " must extend another interface");
        }
        else {
            CtClass target = refiner.getSuperclass();
            if (target.getName().equals("java.lang.Object"))
                throw new WeaveException(refiner.getName()
                            + " must not directly extend java.lang.Object");

            return target;
        }
    }

    /**
     * Appends the interfaces and methods declared in the
     * class to its super class.  <code>refiner</code> will be
     * modified; it should be discarded after this method returns.
     *
     * @param refiner      the class
     * @return the modified super class.
     */
    public void weave(CtClass target, CtClass refiner, ClassMap allRefinerNames, boolean privileged)
        throws WeaveException, CannotCompileException, NotFoundException
    {
        boolean isIntf = refiner.isInterface();
        copyFields(target, refiner, privileged);
        CtClass[] interfaces = refiner.getInterfaces();
        int first = isIntf ? 1 : 0;
        for (int i = first; i < interfaces.length; i++)
            if (!target.subtypeOf(interfaces[i]))
                target.addInterface(interfaces[i]);

        if (!isIntf)
            copyConstructors(target, refiner, allRefinerNames);

        copyClassInitializer(target, refiner, allRefinerNames);
        copyMethods(target, refiner, isIntf, allRefinerNames, privileged);
    }

    private void copyFields(CtClass target, CtClass refiner, boolean privileged)
        throws WeaveException, CannotCompileException, NotFoundException
    {
        CtField[] newFields = refiner.getDeclaredFields();
        for (int i = 0; i < newFields.length; i++) {
            CtField f = newFields[i];
            CtField origF = null;
            try {
                origF = target.getField(f.getName());
            }
            catch (NotFoundException e) {
                CtField newF = new CtField(f, target);
                target.addField(newF);
                return;     // not override any superclass field
            }

            /* Defining a field with the same name as the super's field
             * is not allowed for the time being.
             * This is different from the Java's semantics.
             */
            if (!privileged && !origF.visibleFrom(refiner))
                throw new WeaveException(refiner.getName() + " cannot access a field", origF);

            if (!origF.getSignature().equals(f.getSignature()))
                throw new WeaveException("incompatible field type", f);

            if (Modifier.isFinal(origF.getModifiers()))
                throw new WeaveException("cannot update a final field", origF);

            // copy attributes.
            /* If a field overrides its super's field, then the annotations of the former
               field are appended to the super's field.
             */
            FieldInfo info = f.getFieldInfo2();
            FieldInfo origInfo = origF.getFieldInfo2();
            AttributeInfo ai = info.getAttribute(ConstantAttribute.tag);
            if (ai != null)
                origInfo.addAttribute(ai);

            appendAnnotations(true, info, origInfo);
            appendAnnotations(false, info, origInfo);
        }
    }

    private static void appendAnnotations(boolean visible,
                                          FieldInfo src, FieldInfo dest)
    {
        String tag = visible ? AnnotationsAttribute.invisibleTag
                             : AnnotationsAttribute.visibleTag; 
        AnnotationsAttribute srcAttr
            = (AnnotationsAttribute)src.getAttribute(tag);
        if (srcAttr == null)
            return;

        AnnotationsAttribute destAttr
            = (AnnotationsAttribute)dest.getAttribute(tag);
        if (destAttr == null) {
            dest.addAttribute(srcAttr);
            return;
        }

        Annotation[] anno = srcAttr.getAnnotations();
        for (int i = 0; i < anno.length; i++)
            destAttr.addAnnotation(anno[i]);
    }

    /* The refiner must have only a constructor taking no argument.
     * The super() call in that constructor body is eliminated.
     */
    private static void copyConstructors(CtClass target, CtClass refiner, ClassMap allRefiners)
        throws WeaveException, CannotCompileException, NotFoundException
    {
        CtConstructor[] cons = refiner.getDeclaredConstructors();
        if (cons.length != 1)
            throw new WeaveException("more than one constructors in "
                                     + refiner.getName());

        CtConstructor cc = cons[0];
        if (cc.isEmpty())       // defualt constructor
            return;

        if (!cc.getSignature().equals("()V"))
            throw new WeaveException("constructor with parameters in " + refiner.getName());

        String methodName = target.makeUniqueName("_init");
        CtMethod mth = cc.toMethod(methodName, target, allRefiners);
        mth.setModifiers(Modifier.PRIVATE);
        target.addMethod(mth);

        String flagName = target.makeUniqueName("_init_flag");
        CtField flag = new CtField(CtClass.booleanType, flagName, target);
        target.addField(flag);
        cons = target.getDeclaredConstructors();
        String expr = "if(!" + flagName + "){" + flagName + "=true;" + methodName + "();}";
        for (int i = 0; i < cons.length; i++) {
            CtConstructor c = cons[i];
            c.insertAfter(expr);
        }
    }

    private static void copyClassInitializer(CtClass target, CtClass refiner, ClassMap allRefiners)
        throws WeaveException, CannotCompileException, NotFoundException
    {
        CtConstructor cons = refiner.getClassInitializer();
        if (cons == null || cons.isEmpty())
            return;

        CtConstructor dest = target.getClassInitializer();
        if (dest == null) {
            dest = new CtConstructor(cons, target, allRefiners);
            target.addConstructor(dest);
        }
        else {
            String methodName = target.makeUniqueName("_clinit");
            CtMethod mth = cons.toMethod(methodName, target, allRefiners);
            mth.setModifiers(Modifier.PRIVATE | Modifier.STATIC);
            target.addMethod(mth);
            dest.insertAfter(methodName + "();");
        }
    }

    /**
     * @param allRefiners           this is modified by copyMethods().
     */
    private void copyMethods(CtClass target, CtClass refiner, boolean isIntf,
                             ClassMap allRefiners, boolean privileged)
        throws WeaveException, NotFoundException, CannotCompileException
    {
        /* Overriding an invisible method is not allowed unless privileged is on.
         * For example,
         *
         * public class foo.A {
         *   int foo() { return 0; }  // package only
         * }
         * public class bar.B {
         *   public int foo() { return 0; }
         * }
         *
         * foo() in B does not override foo() in A.  It would be OK
         * to make the two foo()s coexist in A after weaving but
         * that design would be also error prone. 
         */
        HashMap hash = new HashMap();
        CtMethod[] origMethods = target.getDeclaredMethods();
        for (int i = 0; i < origMethods.length; i++) {
            CtMethod m = origMethods[i];
            hash.put(m, m);
        }

        CtMethod[] newMethods = refiner.getDeclaredMethods();
        ArrayList superOrigPairs = null;
        if (privileged)
            superOrigPairs = findSuperMethods(target, refiner, newMethods);

        CtMethod[] addedMethods = new CtMethod[newMethods.length];
        int numAdded = 0;
        CodeConverter refinerConv = new CodeConverter();
        for (int i = 0; i < newMethods.length; i++) {
            CtMethod m = newMethods[i];
            if (!isIntf && Modifier.isAbstract(m.getModifiers()))
                if (privileged && m.getName().startsWith(superPrefix))
                    continue;
                else if (!Modifier.isAbstract(target.getModifiers()))
                    throw new WeaveException(m.getName() + " method in "
                                             + refiner.getName() + " is abstarct");

            CtMethod found = (CtMethod)hash.get(m);
            if (found != null) {
            	if (!privileged && !found.visibleFrom(refiner)) 
            		throw new WeaveException("cannot override an invisible method: "
                            + found.getName() + " in " + target.getName());

                if (!isIntf) {
                    renameOrigMethod(target, found, refinerConv);
                    addedMethods[numAdded++] = m;
                }
            }
            else
                addedMethods[numAdded++] = m;
        }

        if (superOrigPairs != null)
            convSuperToOrig(refinerConv, superOrigPairs);

        allRefiners.fix(target);
        for (int i = 0; i < numAdded; i++) {
            CtMethod m = addedMethods[i];
            m.instrument(refinerConv);          // target should be detached.
            CtMethod newM = CtNewMethod.copy(m, target, allRefiners);
            target.addMethod(newM);
        }
    }

    private static final String superPrefix = "super_";

    private ArrayList findSuperMethods(CtClass target, CtClass refiner, CtMethod[] newMethods)
        throws WeaveException
    {
        ArrayList superOrigPairs = new ArrayList();
        for (int j = 0; j < newMethods.length; j++) {
            CtMethod m = newMethods[j];
            String name = m.getName();
            if (name.startsWith(superPrefix)) {
                String origName = name.substring(superPrefix.length());
                try {
                    CtMethod origM = target.getMethod(origName, m.getSignature());
                    if (origM.getDeclaringClass() != target)
                        throwBadSuper(name, refiner);

                    superOrigPairs.add(m);
                    superOrigPairs.add(origM);
                }
                catch (NotFoundException e) {
                    throwBadSuper(name, refiner);
                }
            }
        }

        return superOrigPairs;
    }

    private void throwBadSuper(String name, CtClass refiner) throws WeaveException {
        throw new WeaveException("bad " + superPrefix + ".. method: "
                                 + name + " in " + refiner.getName());
    }

    private void convSuperToOrig(CodeConverter conv, ArrayList superOrigPairs)
        throws CannotCompileException
    {
        int n = superOrigPairs.size();
        int i = 0;
        while (i < n) {
            CtMethod superM = (CtMethod)superOrigPairs.get(i++);
            CtMethod origM = (CtMethod)superOrigPairs.get(i++);
            conv.redirectMethodCall(superM, origM);
        }
    }

    private static final String origPrefix = "_orig_";

    private void renameOrigMethod(CtClass orig, CtMethod origM, CodeConverter conv)
        throws CannotCompileException
    {
        String origName = origM.getName();
        String methodName = origPrefix + origName;
        CtMethod conflict;
        try {
            conflict = orig.getMethod(methodName, origM.getSignature());
            // methodName is already used.
            methodName = orig.makeUniqueName(methodName);
        }
        catch (NotFoundException e) {}
        origM.setName(methodName);
        int mod = origM.getModifiers();
        origM.setModifiers(Modifier.setPrivate(mod));
        conv.redirectMethodCall(origName, origM);
    }
}
