/*
 * GluonJ, An Aspect-Oriented Programming System for Java.
 * Copyright (C) 2006 Shigeru Chiba. All Rights Reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License.  Alternatively, the contents of this file may be used under
 * the terms of the GNU Lesser General Public License Version 2.1 or later.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 */
package javassist.gluonj.weave;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;

import javassist.CtClass;
import javassist.CannotCompileException;
import javassist.NotFoundException;
import javassist.gluonj.Gluon;
import javassist.gluonj.WeaveException;

/**
 * Another load-time weaver.  Provided only for testing purpose.
 * The standard load-time weaver is javassist.gluonj.weave.HotSwapper.
 *
 * @see HotSwapper
 */
public class RuntimeWeaver extends Weaver {
    private ClassLoader classLoader;
    private boolean running;

    /**
     * Constructs a weaver.
     *
     * @param g                 glue object.
     * @param mainClass         the woven classes are loaded by the same
     *                          class loader as this <code>mainClass</code>.  
     */
    public RuntimeWeaver(Gluon g, Class mainClass) throws WeaveException {
        super(g);
        classLoader = mainClass.getClassLoader();
        running = false;
    }

    /**
     * Executes weaving.  This transforms only the classes that refine
     * classes are applied to or the super classes of those classes.
     */
    public void run() throws RuntimeException {
        running = true;
        Set set = refineTargets.entrySet();
        int size = set.size();
        String[] targetNames = new String[size];
        LinkedList[] refines = new LinkedList[size];
        Iterator it = set.iterator();
        int i = 0;
        while (it.hasNext()) {
            Map.Entry e =(Map.Entry)it.next();
            targetNames[i] = (String)e.getKey();
            refines[i] = (LinkedList)e.getValue();
            i++;
        }

        try {
            for (int j = 0; j < size; j++)
                transform(targetNames[j], refines[j], true);
        }
        catch (WeaveException e) {
            throw new RuntimeException(e.getMessage(), e);
        }
    }

    /**
     * Weaves advices and a given class.
     *
     * @param className         the fully-qualified name of the class.
     */
    public void use(String className) throws RuntimeException {
        if (!running)
            run();

        try {
            transform(className);
        }
        catch (WeaveException err) {
            throw new RuntimeException(err.getMessage(), err);
        }
    }

    protected boolean transform(String className, LinkedList refines,
                                boolean inspect)
        throws WeaveException
    {
        if (className.startsWith("java.") || className.startsWith("javax."))
            return false;     // cannot process basic classes.

        try {
            CtClass target = pool.get(className);
            if (target.isFrozen())
                return false;

            CtClass superClass = target.getSuperclass();
            if (superClass != null)
                transform(superClass.getName(), inspect);

            return super.transform(className, refines, inspect);
        }
        catch (NotFoundException nfe) {
            throw new WeaveException(nfe);
        }
    }

    protected void save(CtClass clazz, boolean modify) throws CannotCompileException {
        try {
            clazz.toClass(classLoader, null);
        }
        catch (CannotCompileException e) {
            Throwable cause = e.getCause();
            if (cause != null && (cause instanceof LinkageError))
                throw (LinkageError)cause;
            else
                throw e;
        }
    }
}
